using System;
using System.Reflection;
using System.Collections.Generic;

namespace MyData
{
    public static class BusinessDataStorage
    {
        private static BusinessObjectCategoryCollection _categories;
        private static BusinessObjectCollection _data;

        public static void Create()
        {
            _categories = new BusinessObjectCategoryCollection();
            _data = new BusinessObjectCollection();
        }

        public static void Create(int objectCount, int categoryCount)
        {
            _categories = new BusinessObjectCategoryCollection(categoryCount);
            _data = new BusinessObjectCollection(objectCount, _categories);
        }

        public static BusinessObjectCollection GetData()
        {
            if (_data == null)
            {
                _data = new BusinessObjectCollection(20, _categories);
            }

            return _data;
        }

        public static BusinessObjectCollection GetData(int categoryId)
        {
            List<BusinessObject> result = _data.FindAll(delegate(BusinessObject obj) { return obj.CategoryID == categoryId; });
            BusinessObjectCollection ret = new BusinessObjectCollection();
            ret.Clear();
            ret.AddRange(result);

            return ret;
        }

        public static int GetCount()
        {
            return GetData().Count;
        }

        public static int GetCategoryCount()
        {
            return GetCategories().Count;
        }

        public static BusinessObjectCategoryCollection GetCategories()
        {
            if (_categories == null)
            {
                _categories = new BusinessObjectCategoryCollection(10, GetData());
            }

            return _categories;
        }

        public static int GetNextID()
        {
            return _data[_data.Count - 1].ID + 1;
        }

        public static BusinessObject Find(BusinessObject obj)
        {
            int index = _data.BinarySearch(obj);
            if (index >= 0 && index < _data.Count)
            {
                return _data[index];
            }

            return null;
        }

        public static BusinessObject Find(int id)
        {
            return Find(new BusinessObject(id));
        }

        public static void Insert(BusinessObject obj)
        {
            BusinessObject insertedObj = new BusinessObject(GetNextID()); ;


            PropertyInfo[] properties = obj.GetType().GetProperties();
            foreach (PropertyInfo prop in properties)
            {
                if (prop.Name != "ID")
                {
                    insertedObj.GetType().GetProperty(prop.Name).SetValue(insertedObj, prop.GetValue(obj, null), null);
                }
            }

            _data.Add(insertedObj);
        }

        public static void Update(BusinessObject obj)
        {
            if (Find(obj) == null)
            {
                throw new NullReferenceException("Object with the specified ID does not exist in store");
            }

            _data[_data.BinarySearch(obj)] = obj;
        }

        public static void Delete(int id)
        {
            BusinessObject deletedObj = Find(id);
            if (deletedObj == null)
            {
                throw new NullReferenceException("Object with the specified ID does not exist in store");
            }

            _data.Remove(deletedObj);
        }
    }

    public class BusinessObjectCollection : List<BusinessObject>
    {
        public BusinessObjectCollection()
            : this(20)
        {

        }

        public BusinessObjectCollection(int itemCount)
            : this(itemCount, null)
        {

        }

        public BusinessObjectCollection(int itemCount, BusinessObjectCategoryCollection categories)
        {
            List<int> categoryIds = new List<int>();
            if (categories != null)
            {
                foreach (BusinessObjectCategory cat in categories)
                {
                    categoryIds.Add(cat.ID);
                }

                Random rand = new Random();

                for (int i = 0; i < itemCount; i++)
                {
                    this.Add(new BusinessObject(i, categoryIds[rand.Next(categoryIds.Count)]));
                }
            }
            else
            {
                for (int i = 0; i < itemCount; i++)
                {
                    this.Add(new BusinessObject(i));
                }
            }
        }
    }

    public class BusinessObjectCategoryCollection : List<BusinessObjectCategory>
    {
        public BusinessObjectCategoryCollection()
            : this(10)
        {

        }

        public BusinessObjectCategoryCollection(int itemCount)
            : this(itemCount, null)
        {

        }

        public BusinessObjectCategoryCollection(int itemCount, BusinessObjectCollection dataInCategories)
        {
            CreateCollection(itemCount);

            if (dataInCategories != null)
            {
                Random rand = new Random();

                foreach (BusinessObject obj in dataInCategories)
                {
                    obj.CategoryID = this[rand.Next(this.Count)].ID;
                }
            }
        }

        private void CreateCollection(int count)
        {
            for (int i = 0; i < count; i++)
            {
                this.Add(new BusinessObjectCategory(i + 1));
            }
        }
    }

    public class BusinessObjectCategory : IComparable<BusinessObjectCategory>
    {
        private int _id;
        private string _name;
        private string _description;

        private static int _nextId = 1;

        public BusinessObjectCategory()
            : this(_nextId++)
        {

        }

        public BusinessObjectCategory(int id)
        {
            _id = id;
            _name = "Business Object Category: " + id.ToString();
            _description = "Description for Business Object Category: " + id.ToString();
        }

        public int ID
        {
            get { return _id; }
        }

        public string Name
        {
            get { return _name; }
            set { _name = value; }
        }

        public string Description
        {
            get { return _description; }
            set { _description = value; }
        }

        #region IComparable<BusinessObjectCategory> Members

        public int CompareTo(BusinessObjectCategory other)
        {
            return this.ID.CompareTo(other.ID);
        }

        #endregion
    }

    public class BusinessObject : IComparable
    {
        private int _id;
        private string _name;
        private int _categoryId;
        private DateTime _date;
        private int _quantity;
        private double _price;
        private bool _available;

        private static int _nextId;

        static Random rand = new Random();

        public BusinessObject()
            : this(_nextId++)
        {
        }

        public BusinessObject(int id)
            : this(id, 0)
        {

        }

        public BusinessObject(int id, int categoryId)
        {
            _id = id;
            _name = "Business object ID: " + _id.ToString();
            _categoryId = categoryId;
            _date = DateTime.Today.AddDays(rand.Next(20) - 10).AddHours(rand.Next(24) - 12).AddMinutes(rand.Next(60) - 30);
            _quantity = rand.Next(100);
            _price = Math.Round(rand.NextDouble() * 100, 2);
            _available = rand.Next(1000) % 2 == 0;
        }

        public override bool Equals(object obj)
        {
            return ((BusinessObject)obj).ID.Equals(this.ID);
        }

        public int ID
        {
            get
            {
                return _id;
            }
        }

        public string Name
        {
            get
            {
                return _name;
            }
            set
            {
                _name = value;
            }
        }

        public int CategoryID
        {
            get { return _categoryId; }
            set { _categoryId = value; }
        }

        public DateTime Date
        {
            get
            {
                return _date;
            }
            set
            {
                _date = value;
            }
        }

        public int Quantity
        {
            get
            {
                return _quantity;
            }
            set
            {
                _quantity = value;
            }
        }

        public double Price
        {
            get
            {
                return _price;
            }
            set
            {
                _price = value;
            }
        }

        public bool Available
        {
            get
            {
                return _available;
            }
            set
            {
                _available = value;
            }
        }

        #region IComparable Members

        public int CompareTo(object obj)
        {
            return ((BusinessObject)obj).ID.CompareTo(this.ID);
        }

        #endregion
    }
}